/* 
  Parallel Biofilm Simulation for Multi Processors
  By Shay Ellison
  7/24/9
  Program to accompany module 
  "Biofilms:  United They Stand, Divided They Colonize"
  by Angela B. Shiflet and George W. Shiflet
 */

#include <stdlib.h>
#include <stdio.h>

#define SIMPLE_SPRNG
#define USE_MPI
#include <mpi.h>
#include <sprng.h>

#define ROOT 0

#define TOP_ROW 0
#define BOT_ROW 1
#define MES_TAG 2
#define CHANGED_UP 3
#define CHANGED_DOWN 4

#define max(a, b) ((a) > (b)) ? (a) : (b)

#define MAX_NUTRIENT 1.0
#define CONSUMED 0.1

#define EMPTY 0
#define BACTERIUM 1
#define DEAD 2
#define BORDER 3
#define TAKEN 10

/* struct for storing a grid location and whether the location is on another processor */
typedef struct location {
	int row, col, otherProc;
} GridLocation;

/* return a random number between [0, 1), for the parallel algorithm sprng is used */
double randomReal(void);

/* return a random integer between [0, UpperBound) */
int randomInt(int upperBound);

int detProcessorRank(int procRank, int numProc, int direction);

void print1dMatrix(int *matrix, int rows, int cols);

/* Init all cells in nutrient grid, including boundaries*/
void initNutrientGrid(int rows, int cols, double **nutrientGrid);

/* */
void applyDiffusion(double **oldNutrientGrid, double **newNutrientGrid, double diffusionRate, 
		int rows, int cols);

/* Return new nutrition value of a cell in nutrition matrix */
double diffusion(double rate, double site, double n, double ne, double e, double se, double s, double sw, 
		double w, double nw);

/* Init all cells in bacteria grid, including boundaries */
void initBacteriaGrid(int rows, int cols, int **bacteriaGrid, double probInitBacteria);

/* probGrow helper function, computes sum of nutrients of all living cells */
double sumNutritionBacteria(int **bacteriaGrid, double **nutrientGrid, int rows, int cols);

/* Return p/E nutrition, i.e. the probability that a cell will divide */
double probGrow(int **bacteriaGrid, double **nutrientGrid, double p, int rows, int cols);

/* return location of chosen neighbor to grow into for two processors */
GridLocation parallelPickNeighbor(int row, int col, int rows, int cols, int n, int e, int s, int w);

/* */
void parallelGrow(int **oldBacteriaGrid, int **newBacteriaGrid, double **nutrientGrid, double p, int rows, int cols, int procRank, int numProcs);

/* */
void processParallelGrowingCells(int **bacteriaGrid, int *locations, int changed);

/* */
void consume(double **oldNutrientGrid, double **newNutrientGrid, int **bacteriaGrid, int rows, int cols);

int main(int argc, char *argv[]) {
	int rows, cols, rowsPerProcessor, i, j, t, n, size1d, timeSteps, aboveProc, belowProc;
	int **oldBacteriaGrid, **newBacteriaGrid, *bacGrid1d, *printBacGrid1d;
	double **oldNutrientGrid, **newNutrientGrid, probInitBacteria, diffusionRate, p;
	int rowsProcBoundaries, colsProcBoundaries;
	int procRank, numProc;
	
	/* Information about Biofilm simulation */
	rows = cols = 8;
	timeSteps = 20;
	probInitBacteria = 0.8;
	diffusionRate = 0.1;
	p = 1;
	
	/* MPI_COMM_WORLD is our chosen communication network which includes all processors running, in this case 2
           * we put in under the name comm of type MPI_Comm for ease of use later on */
	MPI_Comm comm = MPI_COMM_WORLD;
	MPI_Status status;
	
	/* MPI Init and Set-up */
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(comm, &procRank);
	MPI_Comm_size(comm, &numProc);
	
	aboveProc = detProcessorRank(procRank, numProc, -1);
	belowProc = detProcessorRank(procRank, numProc, 1);
	
	/* Domain Decomposition -
           * Here we are dividing the domain into rows / numProc rows with cols number of columns
	 * So if our original matrix has 10 rows, each processor will have a domain of 5 rows */
	
	if (rows % numProc != 0) { /* rows not evenly divisible by number of processors */
		/* make numRows divisible by the number of processors */
		rows += numProc - (rows % numProc); 
	}
	
	rowsPerProcessor = rows / numProc;
	
	/* the number of rows and cols including boundaries */
	rowsProcBoundaries = rowsPerProcessor + 2;
	colsProcBoundaries = cols + 2;
	
	/* for printing purposes */
	size1d = rowsPerProcessor * cols;
	bacGrid1d = malloc(size1d * sizeof(int));
	if (procRank == ROOT) {
		printBacGrid1d = malloc(rows * cols * sizeof(int));
	}
	
	/* allocate memory for matrices */
	oldNutrientGrid = malloc(sizeof(double*) * rowsProcBoundaries); 
	newNutrientGrid = malloc(sizeof(double*) * rowsProcBoundaries); 
	oldBacteriaGrid = malloc(sizeof(int*) * rowsProcBoundaries); 
	newBacteriaGrid = malloc(sizeof(int*) * rowsProcBoundaries); 
	for (i = 0; i < rowsProcBoundaries; i++) {
		oldNutrientGrid[i] = malloc(sizeof(double) * colsProcBoundaries);
		newNutrientGrid[i] = malloc(sizeof(double) * colsProcBoundaries);
		oldBacteriaGrid[i] = malloc(sizeof(int) * colsProcBoundaries);
		newBacteriaGrid[i] = malloc(sizeof(int) * colsProcBoundaries);
	}	
	initNutrientGrid(rowsPerProcessor, cols, oldNutrientGrid);
	initBacteriaGrid(rowsPerProcessor, cols, oldBacteriaGrid, probInitBacteria);
	
	/* Simulation */
	for (t = 0; t < timeSteps; t++) {
		/* copy values from oldBacteriaGrid to a 1d matrix to be gathered together by ROOT */
		n = 0;
		for (i = 1; i <= rowsPerProcessor; i++) {
			for (j = 1; j <= cols; j++) {
				bacGrid1d[n++] = oldBacteriaGrid[i][j];
			}
		}
		
		MPI_Gather(bacGrid1d, size1d, MPI_INT, printBacGrid1d, size1d, MPI_INT, ROOT, comm);
		
		if (procRank == ROOT) {
			print1dMatrix(printBacGrid1d, rows, cols);
			printf("\n");
		}
		
		/* ghost cells from neighbors */
		if (procRank % 2 == 1) { /* Odd Processor Rank */
			/* ------------Nutrient Grid--------------- */
			/* Send top row to ABOVE processor's bottom ghost row */
			MPI_Send(&oldNutrientGrid[1][1], cols, MPI_DOUBLE, aboveProc, TOP_ROW, comm);
			/* Receive BELOW processor's top, place in bottom ghost row */
			MPI_Recv(&oldNutrientGrid[rowsPerProcessor + 1][1], cols, MPI_DOUBLE, belowProc, TOP_ROW, comm, &status);
			/* Send bottom row to processor BELOW's top ghost row */
			MPI_Send(&oldNutrientGrid[rowsPerProcessor][1], cols, MPI_DOUBLE, belowProc, BOT_ROW, comm); 
			/* Receive ABOVE processor's bottom row and place in top ghost row */
			MPI_Recv(&oldNutrientGrid[0][1], cols, MPI_DOUBLE, aboveProc, BOT_ROW, comm, &status);
			/* -----------Bacteria Grid---------------- */
			MPI_Send(&oldBacteriaGrid[1][1], cols, MPI_INT, aboveProc, TOP_ROW, comm);
			MPI_Recv(&oldBacteriaGrid[rowsPerProcessor + 1][1], cols, MPI_INT, belowProc, TOP_ROW, comm, &status);
			MPI_Send(&oldBacteriaGrid[rowsPerProcessor][1], cols, MPI_INT, belowProc, BOT_ROW, comm); 
			MPI_Recv(&oldBacteriaGrid[0][1], cols, MPI_INT, aboveProc, BOT_ROW, comm, &status);
		} else { /* Even Processor Rank */
			/* ------------Nutrient Grid--------------- */
			/* Receive BELOW's top row, place in bottom ghost row */
			MPI_Recv(&oldNutrientGrid[rowsPerProcessor + 1][1], cols, MPI_DOUBLE, belowProc, TOP_ROW, comm, &status);
			/* Send top row to ABOVE's bottom ghost row */
			MPI_Send(&oldNutrientGrid[1][1], cols, MPI_DOUBLE, aboveProc, TOP_ROW, comm);
			/* Receive ABOVE's bottom row, place in top ghost row */
			MPI_Recv(&oldNutrientGrid[0][1], cols, MPI_DOUBLE, aboveProc, BOT_ROW, comm, &status);
			/* Send bottom row to BELOW's top ghost row */
			MPI_Send(&oldNutrientGrid[rowsPerProcessor][1], cols, MPI_DOUBLE, belowProc, BOT_ROW, comm); 
			/* -----------Bacteria Grid---------------- */
			MPI_Recv(&oldBacteriaGrid[rowsPerProcessor + 1][1], cols, MPI_INT, belowProc, TOP_ROW, comm, &status);
			MPI_Send(&oldBacteriaGrid[1][1], cols, MPI_INT, aboveProc, TOP_ROW, comm);
			MPI_Recv(&oldBacteriaGrid[0][1], cols, MPI_INT, aboveProc, BOT_ROW, comm, &status);
			MPI_Send(&oldBacteriaGrid[rowsPerProcessor][1], cols, MPI_INT, belowProc, BOT_ROW, comm); 
		}
		/* Diffuse nutrients in oldNutrientGrid */
		applyDiffusion(oldNutrientGrid, newNutrientGrid, diffusionRate, rowsPerProcessor, cols); /* stored in newNutrientGrid */
		/* Grow the bacteria from the oldBacteriaGrid using the newNutrientGrid */
		parallelGrow(oldBacteriaGrid, newBacteriaGrid, newNutrientGrid, p, rowsPerProcessor, cols, procRank, numProc); /* stored in newBacteriaGrid */
		/* Bacteria on newBacterGrid have consumed nutrients in newNutrientGrid, so subtract from nutrients */
		consume(newNutrientGrid, oldNutrientGrid, newBacteriaGrid, rowsPerProcessor, cols); /* stored in oldNutrientGrid */
		
		/* Copy values from newBacteriaGrid into oldBacteriaGrid - already handled above for nutrientGrid */
		for (i = 1; i <= rowsPerProcessor; i++) {
			for (j = 1; j <= cols; j++) {
				oldBacteriaGrid[i][j] = newBacteriaGrid[i][j];
			}
		}
	}
	/*printf("Processor %d done\n", procRank);*/
	MPI_Finalize();
	return 0;
}

double randomReal(void) {
	return sprng();
}

int randomInt(int upperBound) {
	return isprng() % upperBound;
}

int detProcessorRank(int procRank, int numProc, int direction) {
	return (procRank == 0 && direction == -1) ? numProc - 1 : (procRank + direction) % numProc;
}

void print1dMatrix(int *matrix, int rows, int cols) {
	int i, col;
	col = 1;
	for (i = 0; i < rows * cols; i++) {
		printf("%d", matrix[i]);
		if (col++ == cols) {
			printf("\n");
			col = 1;
		}
	}
}

void initNutrientGrid(int rows, int cols, double **nutrientGrid) {
	int i, j;
	int rowsb = rows + 2;
	int colsb = cols + 2;
	// First and last boundary columns
	for (i = 0; i < rowsb; i++) {
		nutrientGrid[i][0] = 0;
		nutrientGrid[i][colsb - 1] = MAX_NUTRIENT;
	}
	// The center lattice
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			nutrientGrid[i][j] = MAX_NUTRIENT;
		}
	}
}

double diffusion(double diffusionRate, double site, double n, double ne, double e, double se, double s, double sw, double w, double nw) {
	return (1 - (8 * diffusionRate)) * site + (diffusionRate * (n + ne + e + se + s + sw + w + nw));
}

void applyDiffusion(double **oldNutrientGrid, double **newNutrientGrid, double diffusionRate, int rows, int cols) {
	int i, j, in, is, je, jw;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			in = i + 1;
			is = i - 1;
			je = j + 1;
			jw = j - 1;			
			newNutrientGrid[i][j] = diffusion(diffusionRate, oldNutrientGrid[i][j], oldNutrientGrid[in][j],
					oldNutrientGrid[in][je], oldNutrientGrid[i][je], oldNutrientGrid[is][je], oldNutrientGrid[is][j],
					oldNutrientGrid[is][jw], oldNutrientGrid[i][jw], oldNutrientGrid[in][jw]);
			
		}
	}
}

void initBacteriaGrid(int rows, int cols, int **bacteriaGrid, double probInitBacteria) {
	int i, j;
	int rowsb = rows + 2;
	int colsb = cols + 2;
	/* First and last boundary columns */
	for (i = 0; i < rowsb; i++) {
		bacteriaGrid[i][0] = BORDER;
		bacteriaGrid[i][colsb - 1] = BORDER;
	}
	/* First column of lattice */
	for (i = 1; i <= rows; i++) {
		if (randomReal() < probInitBacteria) {
			bacteriaGrid[i][1] = BACTERIUM;
		} else {
			bacteriaGrid[i][1] = EMPTY;
		}
	}
	/* Everything else */
	for (i = 1; i <= rows; i++) {
		for (j = 2; j <= cols; j++) {
			bacteriaGrid[i][j] = EMPTY;
		}
	}
	
}

double sumNutritionBacteria(int **bacteriaGrid, double **nutrientGrid, int rows, int cols) {
	int i, j;
	double sum = 0.0;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			if (bacteriaGrid[i][j] == BACTERIUM) {
				sum += nutrientGrid[i][j];
			}
		}
	}
	return sum;
}

double probGrow(int **bacteriaGrid, double **nutrientGrid, double p, int rows, int cols) {
	double sumOfNutrients = sumNutritionBacteria(bacteriaGrid, nutrientGrid, rows, cols);
	return (sumOfNutrients > 0) ? p / sumOfNutrients : 0.0;
}

GridLocation parallelPickNeighbor(int row, int col, int rows, int cols, int n, int e, int s, int w) {
	int i, count;
	count = 0;
	int lst[4] = {n, e, s, w};
	int pos[4] = {0, 0, 0, 0};
	for (i = 0; i < 4; i++) {
		if (lst[i] == EMPTY) {
			++count;
			pos[i] = i;
		}
	}
	GridLocation loc;
	if (count == 0) {
		loc.row = row;
		loc.col = col;
	} else {
		/*int rand = randomInt(count);*/
		int rand = randomInt(count);
		int index = pos[rand];
		if (index == 0) {
			if (row > 1) {
				loc.row = row - 1;
				loc.col = col;
				loc.otherProc = 0;
			} else { /* row is the top row, so north is last row */
				loc.row = rows;
				loc.col = col;
				loc.otherProc = 1;
			}
			/*printf("Grew to North: %d, col: %d on %d\n", loc.row, loc.col, loc.otherProc);*/
		} else if (index == 1) {
			loc.row = row;
			loc.col = col + 1;
			loc.otherProc = 0;
			/*printf("Grew to East: %d, col: %d on %d\n", loc.row, loc.col, loc.otherProc);*/
		} else if (index == 2) {
			if (row < rows) {
				loc.row = row + 1;
				loc.col = col;
				loc.otherProc = 0;
			} else { /* row is the bottom row, so south is top row */
				loc.row = 1;
				loc.col = col;
				loc.otherProc = 1;
			}
			/*printf("Grew to South: %d, col: %d on %d\n", loc.row, loc.col, loc.otherProc);*/
		} else { /* west */
			loc.row = row;
			loc.col = col - 1;
			loc.otherProc = 0;
			/*printf("Grew to West: %d, col: %d on %d\n", loc.row, loc.col, loc.otherProc);*/
		}
	}
	return loc;
}

void parallelGrow(int **oldBacteriaGrid, int **newBacteriaGrid, double **nutrientGrid, double p, int rows, int cols, int procRank, int numProcs) {
	int i, j, in, is, je, jw, changedAbove, changedBelow, indexA, indexB, recChanged, tempSize;
	double prob = probGrow(oldBacteriaGrid, nutrientGrid, p, rows, cols);
	int len = cols * 2; /* total possible size 2 columns */
	int *locsAbove = malloc(sizeof(int) * len); /* store any updated columns */ 
	int *locsBelow = malloc(sizeof(int) * len);
	int *locr = malloc(sizeof(int) * len * 2);
	MPI_Comm comm = MPI_COMM_WORLD;
	MPI_Status status;
	int aboveProc = detProcessorRank(procRank, numProcs, -1);
	int belowProc = detProcessorRank(procRank, numProcs, 1);
	
	changedAbove = changedBelow = 0;
	indexA = indexB = 0;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			if (oldBacteriaGrid[i][j] == BACTERIUM) {
				if (nutrientGrid[i][j] <= 0) {
					newBacteriaGrid[i][j] = DEAD;
				} else { /* try to divide */
					if (randomReal() < prob * nutrientGrid[i][j]) {
						in = i + 1;
						is = i - 1;
						je = j + 1;
						jw = j - 1;	
						GridLocation loc = parallelPickNeighbor(i, j, rows, cols, oldBacteriaGrid[in][j], oldBacteriaGrid[i][je],
								oldBacteriaGrid[is][j], oldBacteriaGrid[i][jw]);
						if (loc.otherProc) { /* on another processor */
							if (loc.row == 1) { /* grew down -- on below processor */
								changedBelow++;
								locsBelow[indexB++] = loc.row;
								locsBelow[indexB++] = loc.col;
							} else { /* grew up -- on above's processor */
								changedAbove++;
								locsAbove[indexA++] = loc.row;
								locsAbove[indexA++] = loc.col;
							}
						} else {
							newBacteriaGrid[loc.row][loc.col] = BACTERIUM;
							oldBacteriaGrid[loc.row][loc.col] = TAKEN;
						}
						newBacteriaGrid[i][j] = BACTERIUM;
					} else {
						newBacteriaGrid[i][j] = BACTERIUM;
					}
				}
				
			} else {
				if (oldBacteriaGrid[i][j] != TAKEN)
					newBacteriaGrid[i][j] = oldBacteriaGrid[i][j];
			}
		}
	}
	recChanged = 0;

	if (procRank % 2 == 1) { /* Odd processor rank */
		/* Send to Above */
		MPI_Send(locsAbove, changedAbove * 2, MPI_INT, aboveProc, CHANGED_UP, comm);
		/* Receive from Below */
		MPI_Recv(locr, len, MPI_INT, belowProc, CHANGED_UP, comm, &status); 
		MPI_Get_count(&status, MPI_INT, &tempSize);
		recChanged += tempSize;
		/* Send to Below */
		MPI_Send(locsBelow, changedBelow * 2, MPI_INT, belowProc, CHANGED_DOWN, comm);
		/* Receive from Above */
		MPI_Recv(locr + recChanged, len, MPI_INT, aboveProc, CHANGED_DOWN, comm, &status); 
		MPI_Get_count(&status, MPI_INT, &tempSize);
		recChanged += tempSize;
	} else { /* even */
		/* Receive from Below */
		MPI_Recv(locr, len, MPI_INT, belowProc, CHANGED_UP, comm, &status);
		MPI_Get_count(&status, MPI_INT, &tempSize);
		recChanged += tempSize;
		/* Send to Above */
		MPI_Send(locsAbove, changedAbove * 2, MPI_INT, aboveProc, CHANGED_UP, comm);
		/* Receive from Above */
		MPI_Recv(locr + recChanged, len, MPI_INT, aboveProc, CHANGED_DOWN, comm, &status);
		MPI_Get_count(&status, MPI_INT, &tempSize);
		recChanged += tempSize;
		/* Send to Below */
		MPI_Send(locsBelow, changedBelow * 2, MPI_INT, belowProc, CHANGED_DOWN, comm);
	}
	processParallelGrowingCells(newBacteriaGrid, locr, recChanged);
}

void processParallelGrowingCells(int **bacteriaGrid, int *locations, int changed) {
	int i, row, col;
	for (i = 0; i < changed; i += 2) {
		row = locations[i];
		col = locations[i + 1];
		bacteriaGrid[row][col] = BACTERIUM;
	}
}

void consume(double **oldNutrientGrid, double **newNutrientGrid, int **bacteriaGrid, int rows, int cols) {
	int i, j;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			if (bacteriaGrid[i][j] == BACTERIUM) {
				newNutrientGrid[i][j] = max(0.0, oldNutrientGrid[i][j] - CONSUMED);
			} else {
				newNutrientGrid[i][j] = oldNutrientGrid[i][j];
			}
		}
	}
}
